@@ -1,6 +1,20 @@
-(* $I1: Unison file synchronizer: src/transfer.ml $ *)
-(* $I2: Last modified by vouillon on Tue, 31 Aug 2004 11:33:38 -0400 $ *)
-(* $I3: Copyright 1999-2004 (see COPYING for details) $ *)
+(* Unison file synchronizer: src/transfer.ml *)
+(* Copyright 1999-2009, Benjamin C. Pierce 
+
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+*)
+
 
 (* rsync compression algorithm
 
@@ -36,13 +50,13 @@
    characters that could not fill a block.  *)
 
 let debug  = Trace.debug "transfer"
-let debugV = Trace.debug "verbose"
+let debugV = Trace.debug "transfer+"
 let debugToken = Trace.debug "rsynctoken"
 let debugLog =   Trace.debug "rsynclog"
 
 open Lwt
 
-type transfer_instruction = string * int * int
+type transfer_instruction = Bytearray.t * int * int
 
 type transmitter = transfer_instruction -> unit Lwt.t
 
@@ -52,16 +66,14 @@ type transmitter = transfer_instruction -> unit Lwt.t
 
 let reallyRead infd buffer pos length =
   let rec read pos length =
-    let n = Unix.read infd buffer pos length in
+    let n = input infd buffer pos length in
     if n = length || n = 0 then pos + n else
     read (pos + n) (length - n)
   in
   read pos length - pos
 
 let rec reallyWrite outfd buffer pos length =
-  let n = Unix.write outfd buffer pos length in
-  if n < length then reallyWrite outfd buffer (pos + n) (length + n)
-
+  output outfd buffer pos length
 
 (*************************************************************************)
 (*                            TOKEN QUEUE                                *)
@@ -88,7 +100,7 @@ let blockSize64 = Int64.of_int blockSize
 let maxQueueSize = 65500
 let maxQueueSizeFS = Uutil.Filesize.ofInt maxQueueSize
 type tokenQueue =
-  { mutable data : string;       (* the queued tokens *)
+  { mutable data : Bytearray.t;  (* the queued tokens *)
     mutable previous : [`Str of int | `Block of int | `None];
                                  (* some informations about the
                                     previous token *)
@@ -105,29 +117,29 @@ let tokenProg t =
 
 let encodeInt3 s pos i =
   assert (i >= 0 && i < 256 * 256 * 256);
-  s.[pos + 0] <- Char.chr ((i lsr 0) land 0xff);
-  s.[pos + 1] <- Char.chr ((i lsr 8) land 0xff);
-  s.[pos + 2] <- Char.chr ((i lsr 16) land 0xff)
+  s.{pos + 0} <- Char.chr ((i lsr 0) land 0xff);
+  s.{pos + 1} <- Char.chr ((i lsr 8) land 0xff);
+  s.{pos + 2} <- Char.chr ((i lsr 16) land 0xff)
 
 let decodeInt3 s pos =
-  (Char.code s.[pos + 0] lsl 0) lor
-  (Char.code s.[pos + 1] lsl 8) lor
-  (Char.code s.[pos + 2] lsl 16)
+  (Char.code s.{pos + 0} lsl 0) lor
+  (Char.code s.{pos + 1} lsl 8) lor
+  (Char.code s.{pos + 2} lsl 16)
 
 let encodeInt2 s pos i =
   assert (i >= 0 && i < 65536);
-  s.[pos + 0] <- Char.chr ((i lsr 0) land 0xff);
-  s.[pos + 1] <- Char.chr ((i lsr 8) land 0xff)
+  s.{pos + 0} <- Char.chr ((i lsr 0) land 0xff);
+  s.{pos + 1} <- Char.chr ((i lsr 8) land 0xff)
 
 let decodeInt2 s pos =
-  (Char.code s.[pos + 0] lsl 0) lor (Char.code s.[pos + 1] lsl 8)
+  (Char.code s.{pos + 0} lsl 0) lor (Char.code s.{pos + 1} lsl 8)
 
 let encodeInt1 s pos i =
   assert (i >= 0 && i < 256);
-  s.[pos + 0] <- Char.chr i
+  s.{pos + 0} <- Char.chr i
 
 let decodeInt1 s pos =
-  Char.code s.[pos + 0]
+  Char.code s.{pos + 0}
 
 (* Transmit the contents of the tokenQueue *)
 let flushQueue q showProgress transmit cond =
@@ -142,34 +154,34 @@ let flushQueue q showProgress transmit cond =
 
 let pushEOF q showProgress transmit =
   flushQueue q showProgress transmit
-    (q.pos + 1 > String.length q.data) >>= (fun () ->
-  q.data.[q.pos] <- 'E';
+    (q.pos + 1 > Bytearray.length q.data) >>= (fun () ->
+  q.data.{q.pos} <- 'E';
   q.pos <- q.pos + 1;
   q.previous <- `None;
   return ())
 
 let pushString q id transmit s pos len =
-  flushQueue q id transmit (q.pos + len + 3 > String.length q.data)
+  flushQueue q id transmit (q.pos + len + 3 > Bytearray.length q.data)
     >>= (fun () ->
-  if q.pos + 3 + len > String.length q.data then begin
+  if q.pos + 3 + len > Bytearray.length q.data then begin
     (* The file is longer than expected, so the string does not fit in
        the buffer *)
     assert (q.pos = 0);
-    q.data <- String.create maxQueueSize
+    q.data <- Bytearray.create maxQueueSize
   end;
-  q.data.[q.pos] <- 'S';
+  q.data.{q.pos} <- 'S';
   encodeInt2 q.data (q.pos + 1) len;
-  assert (q.pos + 3 + len <= String.length q.data);
-  String.blit s pos q.data (q.pos + 3) len;
+  assert (q.pos + 3 + len <= Bytearray.length q.data);
+  Bytearray.blit_from_string s pos q.data (q.pos + 3) len;
   q.pos <- q.pos + len + 3;
   q.prog <- q.prog + len;
   q.previous <- `Str len;
   return ())
 
 let rec growString q id transmit len' s pos len =
-  let l = min (String.length q.data - q.pos) len in
-  String.blit s pos q.data q.pos l;
-  assert (q.data.[q.pos - len' - 3] = 'S');
+  let l = min (Bytearray.length q.data - q.pos) len in
+  Bytearray.blit_from_string s pos q.data q.pos l;
+  assert (q.data.{q.pos - len' - 3} = 'S');
   assert (decodeInt2 q.data (q.pos - len' - 2) = len');
   let len'' = len' + l in
   encodeInt2 q.data (q.pos - len' - 2) len'';
@@ -182,8 +194,8 @@ let rec growString q id transmit len' s pos len =
     return ()
 
 let pushBlock q id transmit pos =
-  flushQueue q id transmit (q.pos + 5 > String.length q.data) >>= (fun () ->
-  q.data.[q.pos] <- 'B';
+  flushQueue q id transmit (q.pos + 5 > Bytearray.length q.data) >>= (fun () ->
+  q.data.{q.pos} <- 'B';
   encodeInt3 q.data (q.pos + 1) pos;
   encodeInt1 q.data (q.pos + 4) 1;
   q.pos <- q.pos + 5;
@@ -193,7 +205,7 @@ let pushBlock q id transmit pos =
 
 let growBlock q id transmit pos =
   let count = decodeInt1 q.data (q.pos - 1) in
-  assert (q.data.[q.pos - 5] = 'B');
+  assert (q.data.{q.pos - 5} = 'B');
   assert (decodeInt3 q.data (q.pos - 4) + count = pos);
   assert (count < 255);
   encodeInt1 q.data (q.pos - 1) (count + 1);
@@ -222,7 +234,7 @@ let makeQueue length =
       (* We need to make sure here that the size of the queue is not
          larger than 65538
          (1 byte: header, 2 bytes: string size, 65535 bytes: string) *)
-      String.create
+      Bytearray.create
         (if length > maxQueueSizeFS then maxQueueSize else
          Uutil.Filesize.toInt length + 10);
     pos = 0; previous = `None; prog = 0 }
@@ -260,12 +272,12 @@ let send infd length showProgress transmit =
 
 let rec receiveRec outfd showProgress data pos maxPos =
   if pos = maxPos then false else
-  match data.[pos] with
+  match data.{pos} with
     'S' ->
       let length = decodeInt2 data (pos + 1) in
       if Trace.enabled "generic" then debug (fun() -> Util.msg
           "receiving %d bytes\n" length);
-      reallyWrite outfd data (pos + 3) length;
+      reallyWrite outfd (Bytearray.sub data (pos + 3) length) 0 length;
       showProgress length;
       receiveRec outfd showProgress data (pos + length + 3) maxPos
   | 'E' ->
@@ -295,8 +307,7 @@ struct
   let aboveRsyncThreshold sz = sz >= blockSizeFs
 
   (* The type of the info that will be sent to the source host *)
-  type rsync_block_info =
-      { blockSize : int; signatures : (Checksum.t * Digest.t) list }
+  type rsync_block_info = (Checksum.t * Digest.t) list
 
 
   (*** PREPROCESS ***)
@@ -308,7 +319,7 @@ struct
      'blockSize') of the input stream (pointed by 'infd').
      The procedure uses a buffer of size 'bufferSize' to load the input,
      and eventually handles the buffer update. *)
-  let blockIter infd blockSize f arg maxCount =
+  let blockIter infd f arg maxCount =
     let bufferSize = 8192 + blockSize in
     let buffer = String.create bufferSize in
     let rec iter count arg offset length =
@@ -321,7 +332,7 @@ struct
           String.blit buffer offset buffer 0 chunkSize;
           iter count arg 0 chunkSize
         end else begin
-          let l = Unix.read infd buffer length (bufferSize - length) in
+          let l = input infd buffer length (bufferSize - length) in
           if l = 0 then
             arg
           else
@@ -331,6 +342,16 @@ struct
     in
     iter 0 arg 0 0
 
+  let rec rev_split_rec accu1 accu2 n l =
+    if n = 100000 then
+      rev_split_rec (accu2 :: accu1) [] 0 l
+    else
+      match l with
+        []     -> accu2 :: accu1
+      | x :: r -> rev_split_rec accu1 (x :: accu2) (n + 1) r
+
+  let rev_split l = rev_split_rec [] [] 0 l
+
   (* Given a block size, get blocks from the old file and compute a
      checksum and a fingerprint for each one. *)
   let rsyncPreprocess infd =
@@ -344,14 +365,14 @@ struct
     in
     (* Make sure we are at the beginning of the file
        (important for AppleDouble files *)
-    ignore (Unix.lseek infd 0 Unix.SEEK_SET);
+    LargeFile.seek_in infd 0L;
     (* Limit the number of block so that there is no overflow in
        encodeInt3 *)
-    let rev_bi = blockIter infd blockSize addBlock [] (256*256*256) in
-    let bi = Safelist.rev rev_bi in
+    let rev_bi = blockIter infd addBlock [] (256*256*256) in
+    let bi = rev_split rev_bi in
     debugLog (fun() -> Util.msg "%d blocks\n" (Safelist.length bi));
     Trace.showTimer timer;
-    { blockSize = blockSize; signatures = bi }
+    bi
 
 
   (*** DECOMPRESSION ***)
@@ -374,8 +395,7 @@ struct
         reallyWrite outfd decomprBuf 0 length
     in
     let copyBlocks n k =
-      ignore
-        (Unix.LargeFile.lseek infd (Int64.mul n blockSize64) Unix.SEEK_SET);
+      LargeFile.seek_in infd (Int64.mul n blockSize64);
       let length = k * blockSize in
       copy length;
       progress := !progress + length
@@ -383,13 +403,13 @@ struct
     let maxPos = pos + len in
     let rec decode pos =
       if pos = maxPos then false else
-      match data.[pos] with
+      match data.{pos} with
         'S' ->
           let length = decodeInt2 data (pos + 1) in
           if Trace.enabled "rsynctoken" then
             debugToken (fun() ->
               Util.msg "decompressing string (%d bytes)\n" length);
-          reallyWrite outfd data (pos + 3) length;
+          reallyWrite outfd (Bytearray.sub data (pos + 3) length) 0 length;
           progress := !progress + length;
           decode (pos + length + 3)
       | 'B' ->
@@ -421,6 +441,11 @@ struct
 
   let hash checksum = checksum
 
+  let rec sigLength sigs =
+    match sigs with
+      []     -> 0
+    | x :: r -> Safelist.length x + sigLength r
+
   (* Compute the hash table length as a function of the number of blocks *)
   let hashTableLength signatures =
     let rec upperPowerOfTwo n n2 =
@@ -429,19 +454,28 @@ struct
       else
         upperPowerOfTwo n (2 * n2)
     in
-    2 * (upperPowerOfTwo (Safelist.length signatures) 32)
+    2 * (upperPowerOfTwo (sigLength signatures) 32)
 
   (* Hash the block signatures into the hash table *)
   let hashSig hashTableLength signatures =
-    let rec addList hashTable k = function
-      | [] -> hashTable
-      | (cs, fp) :: r ->
+    let hashTable = Array.make hashTableLength [] in
+    let rec addList k l l' =
+      match l, l' with
+        [], [] ->
+          ()
+      | [], r :: r' ->
+          addList k r r'
+      | ((cs, fp) :: r), _ ->
+          (* Negative 31-bits integers are sign-extended when
+             unmarshalled on a 64-bit architecture, so we
+             truncate them back to 31 bits. *)
+          let cs = cs land 0x7fffffff in
           let h = (hash cs) land (hashTableLength - 1) in
           hashTable.(h) <- (k, cs, fp)::(hashTable.(h));
-          addList hashTable (k + 1) r
+          addList (k + 1) r l'
     in
-    let hashTable = Array.make hashTableLength [] in
-    addList hashTable 0 signatures
+    addList 0 [] signatures;
+    hashTable
 
   (* Given a key, retrieve the corresponding entry in the table *)
   let findEntry hashTable hashTableLength checksum :
@@ -506,9 +540,7 @@ struct
   let comprBufSizeFS = Uutil.Filesize.ofInt 8192
 
   (* Compress the file using the algorithm described in the header *)
-  let rsyncCompress
-      { blockSize = blockSize; signatures = sigs }
-      infd srcLength showProgress transmit =
+  let rsyncCompress sigs infd srcLength showProgress transmit =
     debug (fun() -> Util.msg "compressing\n");
     debugLog (fun() -> Util.msg
         "compression buffer size = %d bytes\n" comprBufSize);
@@ -576,7 +608,6 @@ struct
         computeChecksum newOffset toBeSent length miss
       else if length = comprBufSize then begin
         transmitString toBeSent newOffset >>= (fun () ->
-        let toBeSent = newOffset in
         let chunkSize = length - newOffset in
         if chunkSize > 0 then begin
           assert(comprBufSize >= blockSize);
@@ -615,7 +646,7 @@ struct
 
     (* Try to match the current block with one existing in the old file *)
     and processBlock offset toBeSent length checksum =
-      if Trace.enabled "verbose" then
+      if Trace.enabled "transfer+" then
         debugV (fun() -> Util.msg
             "processBlock offset=%d toBeSent=%d length=%d blockSize = %d\n"
             offset toBeSent length blockSize);
